"""Storage endpoints."""

import enum

from .exceptions import BunnyException


class StorageEndpoint(enum.Enum):
    """Represents the storage endpoint which may be used."""

    FALKENSTEIN = "storage.bunnycdn.com"
    NEW_YORK = "ny.storage.bunnycdn.com"
    LOS_ANGELES = "la.storage.bunnycdn.com"
    SINGAPORE = "sg.storage.bunnycdn.com"
    SYDNEY = "syd.storage.bunnycdn.com"

    @staticmethod
    def from_name(name: str) -> "StorageEndpoint":
        """Convert a region code to a StorageEndpoint.

        :param name: The region code/name

        :raises BunnyException: If the name doesn't match anything known.

        :returns: A storage endpoint
        """
        match name:
            case "DE":
                return StorageEndpoint.FALKENSTEIN
            case "NY":
                return StorageEndpoint.NEW_YORK
            case "LA":
                return StorageEndpoint.LOS_ANGELES
            case "SG":
                return StorageEndpoint.SINGAPORE
            case "SYD":
                return StorageEndpoint.SYDNEY

        raise BunnyException(f"Unknown storage endpoint: '{name}'")



import deserialize

from .edge_rule_trigger_type import EdgeRuleTriggerType
from .pattern_matching_type import PatternMatchingType


@deserialize.key("trigger_type", "Type")
@deserialize.auto_snake()
class EdgeRuleTrigger:
    # The unique GUID of the edge rule
    trigger_type: EdgeRuleTriggerType

    # The list of pattern matches that will trigger the edge rule
    pattern_matches: list[str] | None

    # The type of pattern matching
    pattern_matching_type: PatternMatchingType

    # The trigger parameter 1. The value depends on the type of trigger.
    parameter_1: str | None



import datetime

import deserialize

from .pull_zone import PullZone


@deserialize.key("identifier", "id")
@deserialize.parser("date_modified", datetime.datetime.fromisoformat)
@deserialize.auto_snake()
class StorageZone:
    # The ID of the storage zone
    identifier: int

    # The ID of the user that owns the storage zone
    user_id: str | None

    # The name of the storage zone
    name: str | None

    # The API access key or FTP password
    password: str | None

    # The date when the zone was last modified
    date_modified: datetime.datetime

    # Determines if the zone was deleted or not
    deleted: bool

    # The total amount of storage used by this zone
    storage_used: int

    # The total number of files stored by this zone
    files_stored: int

    # The main region used by the storage zone
    region: str | None

    # The replication regions enabled for this storage zone
    replication_regions: list[str] | None

    # The list of pull zones connected to this storage zone
    pull_zones: list[PullZone] | None

    # The read-only API access key or FTP password
    read_only_password: str | None

    # Determines if the storage zone will rewrite 404 status codes to 200 status codes
    rewrite_404_to_200: bool

    # The custom 404 error path that will be returned in case of a missing file
    custom_404_file_path: str | None

    # Determines the storage hostname for this zone
    storage_hostname: str | None

    # Determines the storage zone tier that is storing the data
    zone_tier: int

    # Determines if the storage zone is currently enabling a new replication region
    replication_change_in_progress: bool

    # The custom price override for this zone
    price_override: float  # This should be decimal, but the API returns a float, so we've already lost the information

    # The Storage Zone specific pricing discount
    discount: int



"""Wrapper around the bunny.net pull zone APIs."""

import logging
from typing import Iterator

from bunnynet.httpclient import HttpClient
from bunnynet.models import PullZone


class PullZoneClient:
    """Wrapper class around the bunny.net pull zone APIs."""

    log: logging.Logger
    http_client: HttpClient

    def __init__(
        self,
        *,
        http_client: HttpClient,
        log: logging.Logger,
    ) -> None:
        """Construct a new client object.

        :param http_client: The API HTTP client
        :param log: Any base logger to be used (one will be created if not supplied)
        """

        self.http_client = http_client
        self.log = log.getChild("pullzone")

    def get_all(self, *, include_certificate: bool = False) -> Iterator[PullZone]:
        """Get all storage zones.

        :param include_certificate: Set to true if the result hostnames should contain the SSL certificate.
        """

        parameters = {}

        if include_certificate:
            parameters["includeCertificate"] = True

        yield from self.http_client.get_list("pullzone", PullZone)

    def get(self, identifier: int) -> PullZone | None:
        """Get a zone by its identifier.

        :param identifier: The identifier of the zone to get

        :returns: The zone if found, None otherwise.
        """
        return self.http_client.get(f"pullzone/{identifier}", PullZone)

    def get_by_name(self, name: str) -> PullZone | None:
        """Get a zone by its name.

        :param name: The name of the zone to get

        :returns: The zone if found, None otherwise.
        """
        for zone in self.get_all():
            if zone.name == name:
                return zone
        return None



"""Exceptions."""

import requests


class BunnyException(Exception):
    """Base class for exceptions."""


class BunnyHTTPException(BunnyException):
    """A HTTP exception."""

    response: requests.Response

    def __init__(self, response: requests.Response, message: str = "") -> None:
        super().__init__(message)
        self.response = response



"""Wrapper around the bunny.net storage zone APIs."""

import datetime
import logging
from typing import Iterator

from bunnynet.exceptions import BunnyException
from bunnynet.hashing import sha256 as SHA256
from bunnynet.httpclient import HttpClient
from bunnynet.models import StorageObject, StorageZone
from bunnynet.storage_endpoints import StorageEndpoint


class LogClient:
    """Wrapper class around the bunny.net log APIs."""

    log: logging.Logger
    http_client: HttpClient

    def __init__(
        self,
        *,
        http_client: HttpClient,
        log: logging.Logger,
    ) -> None:
        """Construct a new client object.

        :param http_client: The API HTTP client
        :param log: Any base logger to be used (one will be created if not supplied)
        """

        self.http_client = http_client
        self.log = log.getChild("logs")

    def get(
        self,
        *,
        pull_zone_id: int,
        date: datetime.date,
        status_codes: list[int],
        start_index: int = 0,
        end_index: int = 250,
    ) -> str:
        """Get a zone by its identifier.

        :param pull_zone_id: The identifier of the pull zone to get the logs for
        :param date: The date to get the logs for
        :param status_codes: Only get the logs matching these status codes

        :returns: The logs for the given date
        """

        status_codes_string = ",".join(map(str, status_codes))
        endpoint = (
            str(date.month).zfill(2)
            + "-"
            + str(date.day).zfill(2)
            + "-"
            + str(date.year % 100).zfill(2)
            + "/"
            + str(pull_zone_id)
            + ".log?download=false&status="
            + status_codes_string
            + "&search=&start="
            + str(start_index)
            + "&end="
            + str(end_index)
        )

        return self.http_client.get_raw(endpoint, domain="logging.bunnycdn.com")



import deserialize

from .edge_rule_action_type import EdgeRuleActionType
from .edge_rule_trigger import EdgeRuleTrigger
from .trigger_matching_type import TriggerMatchingType


@deserialize.auto_snake()
class EdgeRule:
    # The unique GUID of the edge rule
    guid: str

    # The action type of the edge rule.
    action_type: EdgeRuleActionType

    # The Action parameter 1. The value depends on other parameters of the edge rule.
    action_parameter_1: str | None

    # The Action parameter 2. The value depends on other parameters of the edge rule.
    action_parameter_2: str | None

    triggers: list[EdgeRuleTrigger]

    # The trigger matching type
    trigger_matching_type: TriggerMatchingType

    # The description of the edge rule
    description: str | None

    # Determines if the edge rule is currently enabled or not
    enabled: bool



import enum


class EdgeRuleTriggerType(enum.Enum):
    URL = 0
    REQUEST_HEADER = 1
    RESPONSE_HEADER = 2
    URL_EXTENSION = 3
    COUNTRY_CODE = 4
    REMOTE_IP = 5
    URL_QUERY_STRING = 6
    RANDOM_CHANCE = 7



"""Wrapper around the bunny.net storage zone APIs."""

import logging
from typing import Iterator

from bunnynet.exceptions import BunnyException
from bunnynet.hashing import sha256 as SHA256
from bunnynet.httpclient import HttpClient
from bunnynet.models import StorageObject, StorageZone
from bunnynet.storage_endpoints import StorageEndpoint


class StorageZoneClient:
    """Wrapper class around the bunny.net storage APIs."""

    log: logging.Logger
    http_client: HttpClient

    def __init__(
        self,
        *,
        http_client: HttpClient,
        log: logging.Logger,
    ) -> None:
        """Construct a new client object.

        :param http_client: The API HTTP client
        :param log: Any base logger to be used (one will be created if not supplied)
        """

        self.http_client = http_client
        self.log = log.getChild("storagezone")

    def get_all(self, *, include_deleted: bool = False) -> Iterator[StorageZone]:
        """Get all storage zones.

        :param include_deleted: Set to true to include deleted storage zones in the query.
        """

        parameters = {}

        if include_deleted:
            parameters["includeDeleted"] = True

        yield from self.http_client.get_list("storagezone", StorageZone, parameters=parameters)

    def get(self, identifier: int) -> StorageZone | None:
        """Get a zone by its identifier.

        :param identifier: The identifier of the zone to get

        :returns: The zone if found, None otherwise.
        """
        return self.http_client.get(f"storagezone/{identifier}", StorageZone)

    def get_by_name(self, name: str) -> StorageZone | None:
        """Get a zone by its name.

        :param name: The name of the zone to get

        :returns: The zone if found, None otherwise.
        """
        for zone in self.get_all():
            if zone.name == name:
                return zone
        return None

    def _resolve_storage_zone(
        self, *, storage_zone_name: str | None = None, storage_zone: StorageZone | None = None
    ) -> StorageZone:
        if storage_zone is not None:
            return storage_zone

        if storage_zone_name is None:
            raise BunnyException(
                "The `storage_zone` and `storage_zone_name` were both None. At least one must be provided."
            )

        storage_zone = self.get_by_name(storage_zone_name)

        if storage_zone is None:
            raise BunnyException(f"Failed to find storage zone named '{storage_zone_name}'")

        return storage_zone

    def list_files(
        self, *, storage_zone_name: str | None = None, storage_zone: StorageZone | None = None, path: str
    ) -> Iterator[StorageObject]:
        """List the files in a storage zone under the given path.

        Either the storage zone or the storage zone name must be provided. The
        former will be used in the case where both are provided.

        :param storage_zone_name: The name of the storage zone to list
        :param storage_zone: The storage zone to liust
        :param path: The path to list under

        :raises BunnyException: If the storage zone is not found, or both storage zone and storage zone name are None.
        """

        storage_zone = self._resolve_storage_zone(storage_zone_name=storage_zone_name, storage_zone=storage_zone)

        yield from self.http_client.get_list(
            endpoint=f"{storage_zone.name}/{path}/",
            response_type=StorageObject,
            domain=StorageEndpoint.from_name(storage_zone.region or "").value,
            access_key=storage_zone.password,
        )

    def upload_file(
        self,
        *,
        storage_zone_name: str | None = None,
        storage_zone: StorageZone | None = None,
        contents: bytes,
        path: str,
        file_name: str,
        sha256: str | None = None,
    ) -> None:
        """Upload a file.

        Either the storage zone or the storage zone name must be provided. The
        former will be used in the case where both are provided.

        :param storage_zone_name: The name of the storage zone to list
        :param storage_zone: The storage zone to liust
        :param contents: The contents of the file
        :param path: The path of the file
        :param file_name: The name of the file
        :param sha256: An optional SHA256 checksum for the API to validate against
        """

        storage_zone = self._resolve_storage_zone(storage_zone_name=storage_zone_name, storage_zone=storage_zone)

        additional_headers = {}

        if sha256:
            additional_headers["Checksum"] = sha256.upper()

        self.http_client.put(
            endpoint=f"{storage_zone.name}/{path}/{file_name}",
            response_type=object,
            body=contents,
            additional_headers=additional_headers,
            domain=StorageEndpoint.from_name(storage_zone.region or "").value,
            access_key=storage_zone.password,
        )

    def upload_text_file(
        self,
        *,
        storage_zone_name: str | None = None,
        storage_zone: StorageZone | None = None,
        contents: str,
        path: str,
        file_name: str,
    ) -> None:
        """Upload a text file.

        Either the storage zone or the storage zone name must be provided. The
        former will be used in the case where both are provided.

        :param storage_zone_name: The name of the storage zone to list
        :param storage_zone: The storage zone to liust
        :param contents: The contents of the text file
        :param path: The path of the file
        :param file_name: The name of the file
        """

        raw_contents = contents.encode("utf-8")

        self.upload_file(
            storage_zone_name=storage_zone_name,
            storage_zone=storage_zone,
            contents=raw_contents,
            path=path,
            file_name=file_name,
            sha256=SHA256(raw_contents),
        )

    def delete_file(
        self,
        *,
        storage_zone_name: str | None = None,
        storage_zone: StorageZone | None = None,
        path: str,
        file_name: str,
    ) -> None:
        """Delete a file.

        Either the storage zone or the storage zone name must be provided. The
        former will be used in the case where both are provided.

        :param storage_zone_name: The name of the storage zone to list
        :param storage_zone: The storage zone to list
        :param path: The path of the file
        :param file_name: The name of the file
        """

        storage_zone = self._resolve_storage_zone(storage_zone_name=storage_zone_name, storage_zone=storage_zone)

        self.http_client.delete(
            endpoint=f"{storage_zone.name}/{path}/{file_name}",
            response_type=object,
            domain=StorageEndpoint.from_name(storage_zone.region or "").value,
            access_key=storage_zone.password,
        )



"""Hash functions"""

import hashlib


def sha256(contents: bytes) -> str:
    """Apply SHA-256 to the supplied contents.

    :param contents: The data to hash

    :returns: The SHA-256 signature of the data
    """

    return hashlib.sha256(contents).hexdigest().upper()



import deserialize


@deserialize.key("identifier", "id")
@deserialize.key("force_ssl", "ForceSSL")
@deserialize.auto_snake()
class Hostname:
    # The unique ID of the hostname
    identifier: int

    # The hostname value for the domain name
    value: str | None

    # Determines if the Force SSL feature is enabled
    force_ssl: bool

    # Determines if this is a system hostname controlled by bunny.net
    is_system_hostname: bool

    # Determines if the hostname has an SSL certificate configured
    has_certificate: bool

    # Contains the Base64Url encoded certificate for the hostname
    certificate: str | None

    # Contains the Base64Url encoded certificate key for the hostname
    certificate_key: str | None



import datetime

import deserialize


def dateparser(value: str) -> datetime.datetime:
    """Parse the date stamps to datetime objects.

    :param value: The raw string

    :returns: The parsed datetime
    """
    date_str, second_fraction = value.split(".")
    timestamp = datetime.datetime.fromisoformat(date_str)
    hundredths = int(second_fraction, base=10)
    timestamp += datetime.timedelta(milliseconds=hundredths * 10)
    return timestamp


@deserialize.parser("last_changed", dateparser)
@deserialize.parser("date_created", dateparser)
@deserialize.auto_snake()
class StorageObject:
    guid: str
    storage_zone_name: str
    path: str
    object_name: str
    length: int
    last_changed: datetime.datetime
    server_id: int
    array_number: int
    is_directory: bool
    user_id: str
    content_type: str
    date_created: datetime.datetime
    storage_zone_id: int
    checksum: str | None
    replicated_zones: str | None  # Comma separated



"""HTTP client for APIs"""

import logging
import urllib.parse
from typing import Any, Iterator, Type, TypeVar, cast

import deserialize
import requests

from bunnynet.exceptions import BunnyHTTPException
from bunnynet.storage_endpoints import StorageEndpoint

BASE_URL = "api.bunny.net"

T = TypeVar("T")


class HttpClient:
    """Base HTTP client for the API."""

    _token: str
    log: logging.Logger

    def __init__(
        self,
        token: str,
        *,
        log: logging.Logger,
    ) -> None:
        """Construct a new client object."""
        self.log = log.getChild("http")
        self._token = token

    def extract_data(self, response: requests.Response) -> Any:
        """Validate a response from the API and extract the data

        :param response: The response to validate

        :raises BunnyHTTPException: On any failure to validate

        :returns: Any data in the response
        """
        _ = self

        if not response.ok:
            raise BunnyHTTPException(response)

        return response.json()

    def get_raw(
        self,
        endpoint: str,
        *,
        attempts: int = 3,
        domain: str = BASE_URL,
    ) -> str:
        """Perform a GET to the endpoint specified.

        :param endpoint: The endpoint to perform the GET on
        :param attempts: Number of attempts remaining to try this call
        :param domain: Override the domain to something else

        :raises BunnyHTTPException: If something goes wrong in a call

        :returns: The response
        """

        url = "https://" + domain + "/" + endpoint

        raw_response = requests.get(
            url,
            headers={
                "AccessKey": self._token,
            },
            timeout=10,
        )

        if not raw_response.ok:
            if attempts > 1 and (raw_response.status_code >= 500):
                return self.get_raw(endpoint, attempts=attempts - 1)

            raise BunnyHTTPException(raw_response, "Failed to get data")

        return raw_response.text

    def get(
        self,
        endpoint: str,
        response_type: Type[T],
        *,
        attempts: int = 3,
        domain: str = BASE_URL,
    ) -> T:
        """Perform a GET to the endpoint specified.

        :param endpoint: The endpoint to perform the GET on
        :param response_type: The type of item the response contains
        :param attempts: Number of attempts remaining to try this call
        :param domain: Override the domain to something else

        :raises BunnyHTTPException: If something goes wrong in a call

        :returns: The response deserialized and cast to the passed in `response_type`
        """

        url = "https://" + domain + "/" + endpoint

        raw_response = requests.get(
            url,
            headers={
                "AccessKey": self._token,
            },
            timeout=10,
        )

        try:
            response_data = self.extract_data(raw_response)
        except BunnyHTTPException as ex:
            if attempts > 1 and (ex.response.status_code >= 500):
                return self.get(endpoint, response_type, attempts=attempts - 1)

            raise

        return cast(T, deserialize.deserialize(response_type, response_data))

    def get_list(
        self,
        endpoint: str,
        response_type: Type[T],
        *,
        page: int = 0,
        attempts: int = 3,
        parameters: dict[str, Any] | None = None,
        domain: str = BASE_URL,
        access_key: str | None = None,
    ) -> Iterator[T]:
        """Perform a GET to the endpoint specified.

        :param endpoint: The endpoint to perform the GET on
        :param response_type: The type of item the response contains
        :param attempts: Number of attempts remaining to try this call
        :param page: The page of results to start from
        :param domain: Override the domain to something else
        :param parameters: A dictionary of parameters to add to the URL
        :param access_key: The access key to use instead of the default

        :raises BunnyHTTPException: If something goes wrong in a call

        :returns: The raw response
        """

        params = urllib.parse.urlencode(parameters or {})

        url = "https://" + domain + "/" + endpoint + f"?page={page}"

        if params:
            url += "&" + params

        raw_response = requests.get(url, headers={"AccessKey": access_key or self._token}, timeout=10)

        try:
            response_data = self.extract_data(raw_response)
        except BunnyHTTPException as ex:
            if attempts > 1 and (ex.response.status_code >= 500):
                yield from self.get_list(endpoint, response_type, page=page, attempts=attempts - 1)
                return

            raise

        deserialized_data = cast(T, deserialize.deserialize(list[response_type], response_data))  # type: ignore

        if isinstance(deserialized_data, list):
            yield from deserialized_data
        else:
            yield deserialized_data

        # TODO
        if isinstance(response_data, dict) and response_data.get("HasMoreItems"):
            yield from self.get_list(endpoint, response_type, page=page + 1)

    def put(
        self,
        endpoint: str,
        response_type: Type[T],
        body: bytes,
        *,
        additional_headers: dict[str, Any] | None = None,
        attempts: int = 3,
        domain: str = BASE_URL,
        access_key: str | None = None,
    ) -> T:
        """Perform a PUT to the endpoint specified.

        :param endpoint: The endpoint to perform the PUT to
        :param response_type: The type of item the response contains
        :param body: The body of the POST message
        :param attempts: Number of attempts remaining to try this call
        :param additional_headers: Any additional headers to add to the call
        :param domain: Override the domain to something else
        :param access_key: The access key to use instead of the default

        :raises BunnyHTTPException: If something goes wrong in a call

        :returns: The response deserialized and cast to the passed in `response_type`
        """

        url = "https://" + domain + "/" + endpoint

        headers = {
            "AccessKey": access_key or self._token,
            "Content-Type": "application/octet-stream",
            "Accept": "application/json",
        }

        if additional_headers:
            headers |= additional_headers

        raw_response = requests.put(url, data=body, headers=headers, timeout=10)

        try:
            response_data = self.extract_data(raw_response)
        except BunnyHTTPException as ex:
            if attempts > 1 and (ex.response.status_code >= 500):
                return self.get(endpoint, response_type, attempts=attempts - 1)

            raise

        if response_type == object:
            return response_data

        return cast(T, deserialize.deserialize(response_type, response_data))

    def delete(
        self,
        endpoint: str,
        response_type: Type[T],
        *,
        additional_headers: dict[str, Any] | None = None,
        attempts: int = 3,
        domain: str = BASE_URL,
        access_key: str | None = None,
    ) -> T:
        """Perform a DELETE to the endpoint specified.

        :param endpoint: The endpoint to perform the PUT to
        :param response_type: The type of item the response contains
        :param attempts: Number of attempts remaining to try this call
        :param additional_headers: Any additional headers to add to the call
        :param domain: Override the domain to something else
        :param access_key: The access key to use instead of the default

        :raises BunnyHTTPException: If something goes wrong in a call

        :returns: The response deserialized and cast to the passed in `response_type`
        """

        url = "https://" + domain + "/" + endpoint

        headers = {
            "AccessKey": access_key or self._token,
            "Content-Type": "application/octet-stream",
            "Accept": "application/json",
        }

        if additional_headers:
            headers |= additional_headers

        raw_response = requests.delete(url, headers=headers, timeout=10)

        try:
            response_data = self.extract_data(raw_response)
        except BunnyHTTPException as ex:
            if attempts > 1 and (ex.response.status_code >= 500):
                return self.get(endpoint, response_type, attempts=attempts - 1)

            raise

        if response_type == object:
            return response_data

        return cast(T, deserialize.deserialize(response_type, response_data))



import datetime

import deserialize

from .bunny_ai_image_blueprint import BunnyAiImageBlueprint
from .edge_rule import EdgeRule
from .hostname import Hostname
from .pull_zone_type import PullZoneType


@deserialize.key("identifier", "id")
@deserialize.key("enable_geo_zone_us", "EnableGeoZoneUS")
@deserialize.key("enable_geo_zone_eu", "EnableGeoZoneEU")
@deserialize.key("enable_geo_zone_asia", "EnableGeoZoneASIA")
@deserialize.key("enable_geo_zone_sa", "EnableGeoZoneSA")
@deserialize.key("enable_geo_zone_af", "EnableGeoZoneAF")
@deserialize.key("pull_zone_type", "type")
@deserialize.key("zone_security_include_hash_remote_ip", "ZoneSecurityIncludeHashRemoteIP")
@deserialize.key("connection_limit_per_ip_count", "ConnectionLimitPerIPCount")
@deserialize.key("enable_webp_vary", "EnableWebPVary")
@deserialize.key("aws_signing_enabled", "AWSSigningEnabled")
@deserialize.key("aws_signing_key", "AWSSigningKey")
@deserialize.key("aws_signing_secret", "AWSSigningSecret")
@deserialize.key("aws_signing_region_name", "AWSSigningRegionName")
@deserialize.key("logging_ip_anonymization_enabled", "LoggingIPAnonymizationEnabled")
@deserialize.key("enable_tls1", "EnableTLS1")
@deserialize.key("enable_tls1_1", "EnableTLS1_1")
@deserialize.key("verify_origin_ssl", "VerifyOriginSSL")
@deserialize.key("optimizer_enable_webp", "OptimizerEnableWebP")
@deserialize.key("optimizer_minify_css", "OptimizerMinifyCSS")
@deserialize.key("optimizer_minify_javascript", "OptimizerMinifyJavaScript")
@deserialize.key("origin_retry_5xx_responses", "OriginRetry5XXResponses")
@deserialize.key("enable_auto_ssl", "EnableAutoSSL")
@deserialize.key("shield_ddos_protection_type", "ShieldDDosProtectionType")
@deserialize.key("shield_ddos_protection_enabled", "ShieldDDosProtectionEnabled")
@deserialize.key("eu_us_discount", "EUUSDiscount")
@deserialize.parser("date_modified", datetime.datetime.fromisoformat)
@deserialize.auto_snake()
class PullZone:
    # The unique ID of the pull zone.
    identifier: int

    # The name of the pull zone.
    name: str | None

    # The origin URL of the pull zone where the files are fetched from.
    origin_url: str | None

    # Determines if the Pull Zone is currently enabled, active and running
    enabled: bool

    # The list of hostnames linked to this Pull Zone
    hostnames: list[Hostname]

    # The ID of the storage zone that the pull zone is linked to
    storage_zone_id: int

    # The ID of the edge script that the pull zone is linked to
    edge_script_id: int

    # The list of referrer hostnames that are allowed to access the pull zone. Requests containing the header Referer: hostname that is not on the list will be rejected. If empty, all the referrers are allowed
    allowed_referrers: list[str] | None

    # The list of referrer hostnames that are allowed to access the pull zone. Requests containing the header Referer: hostname that is not on the list will be rejected. If empty, all the referrers are allowed
    blocked_referrers: list[str] | None

    # The list of IPs that are blocked from accessing the pull zone. Requests coming from the following IPs will be rejected. If empty, all the IPs will be allowed
    blocked_ips: list[str] | None

    # Determines if the delivery from the North American region is enabled for this pull zone
    enable_geo_zone_us: bool

    # Determines if the delivery from the European region is enabled for this pull zone
    enable_geo_zone_eu: bool

    # Determines if the delivery from the Asian / Oceanian region is enabled for this pull zone
    enable_geo_zone_asia: bool

    # Determines if the delivery from the South American region is enabled for this pull zone
    enable_geo_zone_sa: bool

    # Determines if the delivery from the Africa region is enabled for this pull zone
    enable_geo_zone_af: bool

    # True if the URL secure token authentication security is enabled
    zone_security_enabled: bool

    # The security key used for secure URL token authentication
    zone_security_key: str | None

    # True if the zone security hash should include the remote IP
    zone_security_include_hash_remote_ip: bool

    # True if the Pull Zone is ignoring query strings when serving cached objects
    ignore_query_strings: bool

    # The monthly limit of bandwidth in bytes that the pullzone is allowed to use
    monthly_bandwidth_limit: int

    # The amount of bandwidth in bytes that the pull zone used this month
    monthly_bandwidth_used: int

    # The total monthly charges for this so zone so far
    monthly_charges: float

    # Determines if the Pull Zone should forward the current hostname to the origin
    add_host_header: bool

    # Determines the host header that will be sent to the origin
    origin_host_header: str | None

    # The type of the pull zone.
    pull_zone_type: PullZoneType

    # The list of extensions that will return the CORS headers
    access_control_origin_header_extensions: list[str] | None

    # Determines if the CORS headers should be enabled
    enable_access_control_origin_header: bool

    # Determines if the cookies are disabled for the pull zone
    disable_cookies: bool

    # The list of budget redirected countries with the two-letter Alpha2 ISO codes
    budget_redirected_countries: list[str] | None

    # The list of blocked countries with the two-letter Alpha2 ISO codes
    blocked_countries: list[str] | None

    # If true the server will use the origin shield feature
    enable_origin_shield: bool

    # The override cache time for the pull zone
    cache_control_max_age_override: int

    # The override cache time for the pull zone for the end client
    cache_control_public_max_age_override: int

    # Excessive requests are delayed until their number exceeds the maximum burst size.
    burst_size: int

    # Max number of requests per IP per second
    request_limit: int

    # If true, access to root path will return a 403 error
    block_root_path_access: bool

    # If true, POST requests to the zone will be blocked
    block_post_requests: bool

    # The maximum rate at which the zone will transfer data in kb/s. 0 for unlimited
    limit_rate_per_second: float

    # The amount of data after the rate limit will be activated
    limit_rate_after: float

    # The number of connections limited per IP for this zone
    connection_limit_per_ip_count: int

    # The custom price override for this zone
    price_override: float

    # Determines if the Add Canonical Header is enabled for this Pull Zone
    add_canonical_header: bool

    # Determines if the logging is enabled for this Pull Zone
    enable_logging: bool

    # Determines if the cache slice (Optimize for video) feature is enabled for the Pull Zone
    enable_cache_slice: bool

    # Determines if smart caching is enabled for this zone
    enable_smart_cache: bool

    # The list of edge rules on this Pull Zone
    edge_rules: list[EdgeRule]

    # Determines if the WebP Vary feature is enabled.
    enable_webp_vary: bool

    # Determines if the AVIF Vary feature is enabled.
    enable_avif_vary: bool

    # Determines if the Country Code Vary feature is enabled.
    enable_country_code_vary: bool

    # Determines if the Mobile Vary feature is enabled.
    enable_mobile_vary: bool

    # Determines if the Cookie Vary feature is enabled.
    enable_cookie_vary: bool

    # Contains the list of vary parameters that will be used for vary cache by cookie string. If empty, cookie vary will not be used.
    cookie_vary_parameters: list[str] | None

    # Determines if the Hostname Vary feature is enabled.
    enable_hostname_vary: bool

    # The CNAME domain of the pull zone for setting up custom hostnames
    cname_domain: str | None

    # Determines if the AWS Signing is enabled
    aws_signing_enabled: bool

    # The AWS Signing region key
    aws_signing_key: str | None

    # The AWS Signing region secret
    aws_singing_secret: str | None

    # The AWS Signing region name
    aws_signing_region_name: str | None

    # ?
    logging_ip_anonymization_enabled: bool

    # Determines if the TLS 1 is enabled on the Pull Zone
    enable_tls1: bool

    # Determines if the TLS 11.1 is enabled on the Pull Zone
    enable_tls1_1: bool

    # Determines if the Pull Zone should verify the origin SSL certificate
    verify_origin_ssl: bool

    # Determines if custom error page code should be enabled.
    error_page_enable_custom_code: bool

    # Contains the custom error page code that will be returned
    error_page_custom_code: str | None

    # Determines if the statuspage widget should be displayed on the error pages
    error_page_enable_statuspage_widget: bool

    # The statuspage code that will be used to build the status widget
    error_page_statuspage_code: str | None

    # Determines if the error pages should be whitelabel or not
    error_page_whitelabel: bool

    # The zone code of the origin shield
    origin_shield_zone_code: str | None

    # Determines if the log forawrding is enabled
    log_forwarding_enabled: bool

    # The log forwarding hostname
    log_forwarding_hostname: str | None

    # The log forwarding port
    log_forwarding_port: int

    # The log forwarding token value
    log_forwarding_token: str | None

    # Determines the log forwarding protocol type
    log_forwarding_protocol: int

    # Determines if the permanent logging feature is enabled
    logging_save_to_storage: bool

    # The ID of the logging storage zone that is configured for this Pull Zone
    logging_storage_zone_id: int

    # Determines if the zone will follow origin redirects
    follow_redirects: bool

    # The ID of the video library that the zone is linked to
    video_library_id: int

    # The ID of the DNS record tied to this pull zone
    dns_record_id: int

    # The ID of the DNS zone tied to this pull zone
    dns_zone_id: int

    # The cached version of the DNS record value
    dns_record_value: str | None

    # Determines if the optimizer should be enabled for this zone
    optimizer_enabled: bool

    # Determines the maximum automatic image size for desktop clients
    optimizer_desktop_max_width: int

    # Determines the maximum automatic image size for mobile clients
    optimizer_mobile_max_width: int

    # Determines the image quality for desktop clients
    optimizer_image_quality: int

    # Determines the image quality for mobile clients
    optimizer_mobile_image_quality: int

    # Determines if the WebP optimization should be enabled
    optimizer_enable_webp: bool

    # Determines the image manipulation should be enabled
    optimizer_enable_manipulation_engine: bool

    # Determines if the CSS minifcation should be enabled
    optimizer_minify_css: bool

    # Determines if the JavaScript minifcation should be enabled
    optimizer_minify_javascript: bool

    # Determines if image watermarking should be enabled
    optimizer_watermark_enabled: bool

    # Sets the URL of the watermark image
    optimizer_watermark_url: str | None

    # Sets the position of the watermark image
    optimizer_watermark_position: int

    # Sets the offset of the watermark image
    optimizer_watermark_offset: float

    # Sets the minimum image size to which the watermark will be added
    optimizer_watermark_min_image_size: int

    # Determines if the automatic image optimization should be enabled
    optimizer_automatic_optimization_enabled: bool

    # The IP of the storage zone used for Perma-Cache
    perma_cache_storage_zone_id: int

    # The number of retries to the origin server
    origin_retries: int

    # The amount of seconds to wait when connecting to the origin. Otherwise the request will fail or retry.
    origin_connect_timeout: int

    # The amount of seconds to wait when waiting for the origin reply. Otherwise the request will fail or retry.
    origin_response_timeout: int

    # Determines if we should use stale cache while cache is updating
    use_stale_while_updating: bool

    # Determines if we should use stale cache while the origin is offline
    use_stale_while_offline: bool

    # Determines if we should retry the request in case of a 5XX response.
    origin_retry_5xx_responses: bool

    # Determines if we should retry the request in case of a connection timeout.
    origin_retry_connection_timeout: bool

    # Determines if we should retry the request in case of a response timeout.
    origin_retry_response_timeout: bool

    # Determines the amount of time that the CDN should wait before retrying an origin request.
    origin_retry_delay: int

    # Contains the list of vary parameters that will be used for vary cache by query str. If empty, all parameters will be used to construct the key
    querystr_vary_parameters: list[str] | None

    # Determines if the origin shield concurrency limit is enabled.
    origin_shield_enable_concurrency_limit: bool

    # Determines the number of maximum concurrent requests allowed to the origin.
    origin_shield_max_concurrent_requests: int

    # ?
    enable_safe_hop: bool

    # Determines if bunny.net should be caching error responses
    cache_error_responses: bool

    # Determines the max queue wait time
    origin_shield_queue_max_wait_time: int

    # Determines the max number of origin requests that will remain in the queu
    origin_shield_max_queued_requests: int

    # Contains the list of optimizer classes
    optimizer_classes: list[dict] | None

    # Determines if the optimizer class list should be enforced
    optimizer_force_classes: bool

    # Determines if cache update is performed in the background.
    use_background_update: bool

    # If set to true, any hostnames added to this Pull Zone will automatically enable SSL.
    enable_auto_ssl: bool

    # If set to true the query str ordering property is enabled.
    enable_query_string_ordering: bool

    # Gets the log anonymization type for this pull zone
    log_anonymization_type: int

    # 1
    log_format: int

    # 1
    log_forwarding_format: int

    # 1 2
    shield_ddos_protection_type: int

    # ?
    shield_ddos_protection_enabled: bool

    # The type of the origin for this Pull Zone
    origin_type: int

    # Determines if request coalescing is currently enabled.
    enable_request_coalescing: bool

    # Determines the lock time for coalesced requests.
    request_coalescing_timeout: int

    # Returns the link short preview value for the pull zone origin connection.
    origin_link_value: str | None

    # If true, the built-in let's encrypt is disabled and requests are passed to the origin.
    disable_lets_encrypt: bool

    # ?
    enable_bunny_image_ai: bool

    # ?
    bunny_ai_image_blueprints: list[BunnyAiImageBlueprint] | None

    # Determines if the preloading screen is currently enabled
    preloading_screen_enabled: bool

    # The custom preloading screen code
    preloading_screen_code: str | None

    # The preloading screen logo URL
    preloading_screen_logo_url: str | None

    # The currently configured preloading screem theme. (0 - Light, 1 - Dark)
    preloading_screen_theme: int

    # The delay in miliseconds after which the preloading screen will be desplayed
    preloading_screen_delay: int

    # The Pull Zone specific pricing discount for EU and US region.
    eu_us_discount: int

    # The Pull Zone specific pricing discount for South America region.
    south_america_discount: int

    # The Pull Zone specific pricing discount for Africa region.
    africa_discount: int

    # The Pull Zone specific pricing discount for Asia & Oceania region.
    asia_oceania_discount: int

    # The list of routing filters enabled for this zone
    routing_filters: list[str]



import enum


class EdgeRuleActionType(enum.Enum):
    FORCE_SSL = 0
    REDIRECT = 1
    ORIGIN_URL = 2
    OVERRIDE_CACHE_TIME = 3
    BLOCK_REQUEST = 4
    SET_RESPONSE_HEADER = 5
    SET_REQUEST_HEADER = 6
    FORCE_DOWNLOAD = 7
    DISABLE_TOKEN_AUTHENTICATION = 8
    ENABLE_TOKEN_AUTHENTICATION = 9
    OVERRIDE_CACHE_TIME_PUBLIC = 10
    IGNORE_QUERY_STRING = 11
    DISABLE_OPTIMIZER = 12
    FORCE_COMPRESSION = 13
    SET_STATUS_CODE = 14
    BYPASS_PERMA_CACHE = 15
    OVERRIDE_BROWSER_CACHE_TIME = 16



"""Wrapper around the bunny.net APIs."""

import base64
import datetime
import hashlib
import logging
import urllib.parse
from typing import Any

from bunnynet.httpclient import HttpClient
from bunnynet.log_client import LogClient
from bunnynet.pull_zone_client import PullZoneClient
from bunnynet.storage_zone_client import StorageZoneClient


class BunnyClient:
    """Wrapper class around the bunny.net API."""

    log: logging.Logger
    _http_client: HttpClient

    pull_zones: PullZoneClient
    storage_zones: StorageZoneClient
    logs: LogClient

    def __init__(
        self,
        token: str,
        *,
        log: logging.Logger | None = None,
    ) -> None:
        """Construct a new client object.

        :param token: The API token to use.
        :param log: Any base logger to be used (one will be created if not supplied)
        """

        if log is not None:
            self.log = log.getChild("bunnynet")
        else:
            self.log = logging.getLogger("bunnynet")

        self._http_client = HttpClient(token, log=self.log)

        self.pull_zones = PullZoneClient(http_client=self._http_client, log=self.log)
        self.storage_zones = StorageZoneClient(
            http_client=self._http_client,
            log=self.log,
        )
        self.logs = LogClient(http_client=self._http_client, log=self.log)

    @staticmethod
    def get_storage_zone_client(token: str, log: logging.Logger) -> StorageZoneClient:
        """Get a storage zone client directly.

        :param token: The API token for auth
        :param log: The logger to use for debugging

        :returns: A new storage zone client.
        """

        if log is not None:
            log = log.getChild("bunnynet")
        else:
            log = logging.getLogger("bunnynet")

        http_client = HttpClient(token, log=log)

        return StorageZoneClient(http_client=http_client, log=log)

    def generate_url_signature(
        self,
        url: str,
        *,
        path: str | None = None,
        key: str,
        expiration: datetime.datetime,
        token_countries: list[str] | None = None,
        token_countries_blocked: list[str] | None = None,
    ) -> dict[str, Any]:
        """Generate a URL signature parameters so that it has a token which can be used to authenticate it.

        If you pass in a folder with the corresponding path, it will give access
        to that entire folder.

        :param url: The URL to sign. e.g. http://test.b-cdn.net/foo/bar/file.png
        :param path: The path to give access to with this token. e.g. "/foo/bar"
                     If not specified, the token will be valid for the URL only.
        :param key: The signing key for generating signatures
        :param expiration: The datetime that the token should expire
        :param token_countries: If specified, the token is only valid in these countries
        :param token_countries_blocked: If specified, the token will not be valid in these countries.

        :returns: The signature data which can be added to a URL later
        """
        parsed_url = urllib.parse.urlparse(url)
        parameters: dict[str, Any] = urllib.parse.parse_qs(parsed_url.query)

        if path:
            signature_path = path
            parameters["token_path"] = path
        else:
            signature_path = parsed_url.path

        if token_countries:
            parameters["token_countries"] = ",".join(token_countries)

        if token_countries_blocked:
            parameters["token_countries_blocked"] = ",".join(token_countries_blocked)

        parameter_data = "&".join([name + "=" + "".join(value) for name, value in parameters.items()])

        hashable_base = key + signature_path + str(int(expiration.timestamp())) + parameter_data
        token_bytes = base64.b64encode(hashlib.sha256(str.encode(hashable_base)).digest())
        token = token_bytes.decode().replace("\n", "").replace("+", "-").replace("/", "_").replace("=", "")

        result = {"token": token, "expires": int(expiration.timestamp())}

        for name, value in parameters.items():
            result[name] = value

        return result

    def sign_url(
        self,
        url: str,
        *,
        path: str | None = None,
        key: str,
        expiration: datetime.datetime,
        token_countries: list[str] | None = None,
        token_countries_blocked: list[str] | None = None,
    ) -> str:
        """Sign a URL so that it has a token which can be used to authenticate it.

        If you pass in a folder with the corresponding path, it will give access
        to that entire folder.

        :param url: The URL to sign. e.g. http://test.b-cdn.net/foo/bar/file.png
        :param path: The path to give access to with this token. e.g. "/foo/bar"
                     If not specified, the token will be valid for the URL only.
        :param key: The signing key for generating signatures
        :param expiration: The datetime that the token should expire
        :param token_countries: If specified, the token is only valid in these countries
        :param token_countries_blocked: If specified, the token will not be valid in these countries.

        :returns: The full URL with the signature added
        """

        parsed_url = urllib.parse.urlparse(url)

        signature = self.generate_url_signature(
            url,
            path=path,
            key=key,
            expiration=expiration,
            token_countries=token_countries,
            token_countries_blocked=token_countries_blocked,
        )

        return parsed_url.scheme + "://" + parsed_url.netloc + parsed_url.path + "?" + urllib.parse.urlencode(signature)
